from __future__ import annotations

from enum import Enum
from textwrap import dedent
from xml.sax.saxutils import escape


class SoapDocumentType(str, Enum):
    CPF = "CPF"
    CNS = "CNS"


SOAP_TEMPLATE = dedent(
    """\
    <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:urn="urn:ihe:iti:xds-b:2007" xmlns:urn1="urn:oasis:names:tc:ebxml-regrep:xsd:lcm:3.0" xmlns:urn2="urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0" xmlns:urn3="urn:ihe:iti:xds-b:2007">
       <soap:Body>
          <PRPA_IN201305UV02 xsi:schemaLocation="urn:hl7-org:v3 ./schema/HL7V3/NE2008/multicacheschemas/PRPA_IN201305UV02.xsd" ITSVersion="XML_1.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="urn:hl7-org:v3">
             <id root="2.16.840.1.113883.4.714" extension="123456"/>
             <creationTime value="20070428150301"/>
             <interactionId root="2.16.840.1.113883.1.6" extension="PRPA_IN201305UV02"/>
             <processingCode code="T"/>
             <processingModeCode code="T"/>
             <acceptAckCode code="AL"/>
             <receiver typeCode="RCV">
                <device classCode="DEV" determinerCode="INSTANCE">
                   <id root="2.16.840.1.113883.3.72.6.5.100.85"/>
                </device>
             </receiver>
             <sender typeCode="SND">
                <device classCode="DEV" determinerCode="INSTANCE">
                   <id root="2.16.840.1.113883.3.72.6.2"/>
                   <name>{system_code}</name>
                </device>
             </sender>
             <controlActProcess classCode="CACT" moodCode="EVN">
                <code code="PRPA_TE201305UV02" codeSystem="2.16.840.1.113883.1.6"/>
                <queryByParameter>
                   <queryId root="1.2.840.114350.1.13.28.1.18.5.999" extension="1840997084"/>
                   <statusCode code="new"/>
                   <responseModalityCode code="R"/>
                   <responsePriorityCode code="I"/>
                   <parameterList>
                      <livingSubjectId>
                         <value root="{document_root}" extension="{identifier}"/>
                         <semanticsText>LivingSubject.id</semanticsText>
                      </livingSubjectId>
                   </parameterList>
                </queryByParameter>
             </controlActProcess>
          </PRPA_IN201305UV02>
       </soap:Body>
    </soap:Envelope>
    """
).strip()

DOCUMENT_ROOTS = {
    SoapDocumentType.CPF: "2.16.840.1.113883.13.237",
    SoapDocumentType.CNS: "2.16.840.1.113883.13.236",
}


def build_busca_pessoa_envelope(
    identifier: str,
    document_type: SoapDocumentType,
    *,
    system_code: str,
) -> str:
    return SOAP_TEMPLATE.format(
        system_code=escape(system_code),
        document_root=DOCUMENT_ROOTS[document_type],
        identifier=escape(identifier),
    )




from __future__ import annotations

import base64
import json
import time
import unittest

import httpx

from three_s_cadsus_client import CadSUSClient, CadSUSSettings
from three_s_cadsus_client.client import DocumentType, get_document_type, normalize_identifier
from three_s_cadsus_client.config import AuthMethod
from three_s_cadsus_client.soap import SoapDocumentType, build_busca_pessoa_envelope


def build_fake_jwt(expiration: int | None = None) -> str:
    header = base64.urlsafe_b64encode(json.dumps({"alg": "none"}).encode()).decode().rstrip("=")
    payload = {}
    if expiration is not None:
        payload["exp"] = expiration
    body = base64.urlsafe_b64encode(json.dumps(payload).encode()).decode().rstrip("=")
    return f"{header}.{body}.signature"


class CadSUSClientTests(unittest.IsolatedAsyncioTestCase):
    async def test_buscar_pessoa_uses_cached_token(self) -> None:
        counters = {"login": 0, "token": 0, "api": 0}
        jwt_token = build_fake_jwt(int(time.time()) + 3600)

        def handler(request: httpx.Request) -> httpx.Response:
            if request.url == httpx.URL("https://example.test/login"):
                counters["login"] += 1
                return httpx.Response(200, json={"access_token": "login-token"})

            if request.url == httpx.URL("https://example.test/token"):
                counters["token"] += 1
                self.assertEqual(request.headers.get("Authorization"), "Bearer login-token")
                return httpx.Response(200, json={"access_token": jwt_token})

            if request.url == httpx.URL("https://example.test/api"):
                counters["api"] += 1
                self.assertEqual(request.headers.get("Authorization"), f"jwt {jwt_token}")
                return httpx.Response(200, text="<ok/>")

            self.fail(f"Unexpected request URL: {request.url}")

        settings = CadSUSSettings(
            auth_method=AuthMethod.API,
            auth_login_url="https://example.test/login",
            auth_token_url="https://example.test/token",
            api_url="https://example.test/api",
            user="user",
            password="password",
        )

        async with CadSUSClient(
            settings,
            transport=httpx.MockTransport(handler),
        ) as client:
            await client.buscar_pessoa("123.456.789-01")
            await client.buscar_pessoa("12345678901")

        self.assertEqual(counters, {"login": 1, "token": 1, "api": 2})

    async def test_buscar_pessoa_identifies_cns(self) -> None:
        counters = {"token": 0}
        jwt_token = build_fake_jwt(int(time.time()) + 3600)

        def handler(request: httpx.Request) -> httpx.Response:
            if request.url == httpx.URL("https://example.test/token"):
                counters["token"] += 1
                return httpx.Response(200, json={"access_token": jwt_token})
            if request.url == httpx.URL("https://example.test/api"):
                content = request.content.decode()
                self.assertIn('root="2.16.840.1.113883.13.236"', content)
                self.assertIn('extension="898001160366001"', content)
                return httpx.Response(200, text="<ok/>")
            self.fail(f"Unexpected request URL: {request.url}")

        settings = CadSUSSettings(
            auth_method=AuthMethod.CERT,
            auth_token_url="https://example.test/token",
            api_url="https://example.test/api",
            cert="/tmp/cert.pem",
            key="/tmp/key.pem",
        )

        async with CadSUSClient(
            settings,
            transport=httpx.MockTransport(handler),
        ) as client:
            result = await client.buscar_pessoa("898001160366001")

        self.assertEqual(result.document_type, DocumentType.CNS)
        self.assertEqual(counters["token"], 1)


class IdentifierTests(unittest.TestCase):
    def test_normalize_identifier_removes_non_digits(self) -> None:
        self.assertEqual(normalize_identifier("123.456.789-01 "), "12345678901")

    def test_get_document_type_uses_length(self) -> None:
        self.assertEqual(get_document_type("12345678901"), DocumentType.CPF)
        self.assertEqual(get_document_type("898001160366001"), DocumentType.CNS)

    def test_build_busca_pessoa_envelope_uses_correct_root(self) -> None:
        cpf_envelope = build_busca_pessoa_envelope(
            "12345678901",
            SoapDocumentType.CPF,
            system_code="MY-SYSTEM",
        )
        cns_envelope = build_busca_pessoa_envelope(
            "898001160366001",
            SoapDocumentType.CNS,
            system_code="CADSUS",
        )

        self.assertIn('root="2.16.840.1.113883.13.237"', cpf_envelope)
        self.assertIn('extension="12345678901"', cpf_envelope)
        self.assertIn("<name>MY-SYSTEM</name>", cpf_envelope)
        self.assertIn('root="2.16.840.1.113883.13.236"', cns_envelope)
        self.assertIn('extension="898001160366001"', cns_envelope)




from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass
from typing import Protocol

from .config import CadSUSSettings


@dataclass(frozen=True, slots=True)
class CachedToken:
    value: str
    expires_at: float


class TokenCache(Protocol):
    async def get(self, key: str) -> CachedToken | None:
        ...

    async def set(self, key: str, token: CachedToken) -> None:
        ...


_MEMORY_STORE: dict[str, CachedToken] = {}
_MEMORY_LOCK = asyncio.Lock()


class InMemoryTokenCache:
    async def get(self, key: str) -> CachedToken | None:
        async with _MEMORY_LOCK:
            entry = _MEMORY_STORE.get(key)
            if entry is None:
                return None
            if entry.expires_at <= time.time():
                _MEMORY_STORE.pop(key, None)
                return None
            return entry

    async def set(self, key: str, token: CachedToken) -> None:
        async with _MEMORY_LOCK:
            _MEMORY_STORE[key] = token


class DjangoTokenCache:
    def __init__(self, alias: str) -> None:
        from django.core.cache import caches

        self._cache = caches[alias]

    async def get(self, key: str) -> CachedToken | None:
        payload = await asyncio.to_thread(self._cache.get, key)
        if not payload:
            return None
        entry = CachedToken(
            value=payload["value"],
            expires_at=float(payload["expires_at"]),
        )
        if entry.expires_at <= time.time():
            await asyncio.to_thread(self._cache.delete, key)
            return None
        return entry

    async def set(self, key: str, token: CachedToken) -> None:
        ttl = max(1, int(token.expires_at - time.time()))
        payload = {"value": token.value, "expires_at": token.expires_at}
        await asyncio.to_thread(self._cache.set, key, payload, ttl)


def create_token_cache(settings: CadSUSSettings) -> TokenCache:
    try:
        return DjangoTokenCache(settings.cache_alias)
    except Exception:
        return InMemoryTokenCache()

